#ifndef VECTOR_HPP_
#define VECTOR_HPP_

#include <algorithm>
#include <iterator>
#include <stdexcept>

#include "support/concepts.hpp"

template <typename T>
class Vector {

public:
  using value_type = T;
  using size_type = std::size_t;
  using difference_type = std::ptrdiff_t;

  using reference = value_type&;
  using const_reference = value_type const&;

  using pointer = value_type*;
  using const_pointer = value_type const*;

  using iterator = pointer;
  using const_iterator = const_pointer;

  using reverse_iterator = std::reverse_iterator<iterator>;
  using const_reverse_iterator = std::reverse_iterator<const_iterator>;

private:
  template <typename Iter>
  static constexpr bool is_iterator = concepts::requires_<concepts::Iterator, Iter>::value;

public:

  // ////////// CONSTRUCTORS //////////

  // default ctor
  Vector ()
    : m_size (0),
      m_capacity (0),
      m_data (nullptr)
  {
  }

  // size ctor
  Vector (size_type count)
    : m_size (),
      m_capacity (count),
      m_data (new T[m_capacity])
  {
    for (size_type i = 0; i < count; ++i) {
      emplace_back();
    }
  }

  // size w/ default value ctor
  Vector (size_type count, T const& value)
    : m_size (count),
      m_capacity (count),
      m_data (new T[m_capacity])
  {
    std::fill (begin(), end(), value);
  }

  // range ctor
  template <
    typename InputIterator,
    typename = std::enable_if_t<is_iterator<InputIterator>>>
  Vector (InputIterator first, InputIterator last)
    : m_size (0 /* TODO */),
      m_capacity (0 /* TODO */),
      m_data (nullptr /* TODO */)
  {
    // TODO!
    // Initialize a Vrray from the range [first, last).
  }

  // copy ctor
  Vector (Vector const& other)
    : m_size (0 /* TODO */),
      m_capacity (0 /* TODO */),
      m_data (nullptr /* TODO */)
  {
    // TODO!
    // Initialize this object from "other".
  }

  // move ctor
  Vector (Vector&& other) noexcept
    : m_size (std::exchange (other.m_size, 0)),
      m_capacity (std::exchange (other.m_capacity, 0)),
      m_data (std::exchange (other.m_data, nullptr))
  {
  }

  // intializer_list ctor
  Vector (std::initializer_list<T> init)
    : Vector (init.begin(), init.end())
  {
  }

  // ////////// DESTRUCTOR //////////

  // dtor
  ~Vector()
  {
    // TODO!
    // Release allocated memory.
  }

  // ////////// ASSIGNMENT OPERATOR //////////

  // copy assignment
  Vector&
  operator= (Vector const& other)
  {
    // TODO!
    // Assign "other" to this object.
    //   Be careful to check for self-assignment.
  }

  // move assignment
  Vector&
  operator= (Vector&& other)
  {
    m_size = std::exchange (other.m_size, 0);
    m_capacity = std::exchange (other.m_capacity, 0);
    m_data = std::exchange (other.m_data, nullptr);
    return *this;
  }

  // initializer_list assign
  Vector&
  operator= (std::initializer_list<T> ilist)
  {
    // TODO!
    // assign the contents of this vector to be all values
    //   from the passed initializer_list
    // HINT: ensure capacity then copy elements
  }

  // count w/ default value assign
  void
  assign (size_type count, T const& value)
  {
    // TODO!
    // assign the contents of the Vector to be
    //   "count" copies of "value"
  }

  // Range assign
  template <typename InputIterator>
  std::enable_if_t<is_iterator<InputIterator>, void>
  assign (InputIterator first, InputIterator last)
  {
    for (InputIterator i = first; i != last; ++i)
    {
      push_back (*i);
    }
  }

  // initializer_list assign
  void
  assign (std::initializer_list<T> ilist)
  {
    assign (ilist.begin(), ilist.end());
  }

  // ////////// ELEMENT ACCESS //////////

  // bounds-check subscript
  reference
  at (size_type pos)
  {
    if (pos < 0 || pos >= m_size) {
      throw std::out_of_range ("Vector index out-of-bounds");
    }
    return operator[] (pos);
  }

  // bounds-check subscript (const version)
  const_reference
  at (size_type pos) const
  {
    if (pos < 0 || pos >= m_size) {
      throw std::out_of_range ("Vector index out-of-bounds");
    }
    return operator[] (pos);
  }

  // subscript
  reference
  operator[] (size_type pos)
  {
    // TODO!
    // Return the element at position "index".
  }

  // subscript (const version)
  const_reference
  operator[] (size_type pos) const
  {
    // TODO!
    // Return the element at position "index".
  }

  // First value of Vector
  reference
  front()
  {
    return *begin();
  }

  // First value of Vector (const version)
  const_reference
  front() const
  {
    return *begin();
  }

  // Last value of Vector
  reference
  back()
  {
    // TODO!
  }

  // Last value of Vector (const version)
  const_reference
  back() const
  {
    // TODO!
  }

  // Pointer-to-data
  T*
  data() noexcept
  {
    return m_data;
  }

  // Pointer-to-data (const version)
  T const*
  data() const noexcept
  {
    return m_data;
  }

  // ////////// ITERATOR ACCESS //////////

  // Iterator pointing to first element
  iterator
  begin() noexcept
  {
    // TODO!
  }

  // Iterator pointing to first element (const version)
  const_iterator
  begin() const noexcept
  {
    // TODO! -- hint: same as above
  }

  // Iterator pointing to first element (const version)
  const_iterator
  cbegin() const noexcept
  {
    return begin();
  }

  // Iterator pointing to one-past-the-last element
  iterator
  end() noexcept
  {
    // TODO!
  }

  // Iterator pointing to one-past-the-last element (const version)
  const_iterator
  end() const noexcept
  {
    // TODO! -- hint: same as above
  }

  // Iterator pointing to first element (const version)
  const_iterator
  cend() const noexcept
  {
    return end();
  }

  // ////////// REVERSE ITERATOR ACCESS //////////

  reverse_iterator
  rbegin() noexcept
  {
    return {end() - 1};
  }

  const_reverse_iterator
  rbegin() const noexcept
  {
    return {end() - 1};
  }

  const_reverse_iterator
  crbegin() const noexcept
  {
    return {cend() - 1};
  }

  reverse_iterator
  rend() noexcept
  {
    return {begin() - 1};
  }

  const_reverse_iterator
  rend() const noexcept
  {
    return {begin() - 1};
  }

  const_reverse_iterator
  crend() const noexcept
  {
    return {cbegin() - 1};
  }

  // ////////// CAPACITY //////////

  // accessor -- is the Vector empty?
  bool
  empty () const noexcept
  {
    // TODO!
  }

  // accessor -- size of the Vector
  size_type
  size () const noexcept
  {
    return m_size;
  }

  size_type
  max_size () const noexcept
  {
    return std::numeric_limits<size_type>::max();
  }

  // Reserve capacity for "new_cap" elements.
  // "new_cap" must be  greater than capacity.
  //   If not, leave the capacity unchanged.
  // "size" must remain unchanged.
  void
  reserve (size_type new_cap)
  {
    if (new_cap > capacity ())
    {
      m_capacity = new_cap;
      T* data = new T[m_capacity];
      std::copy (begin (), end (), data);
      delete[] std::exchange(m_data, data);
    }
  }

  // accessor -- capacity of the Vector
  size_type
  capacity () const noexcept
  {
    // TODO!
    return 0;
  }

  // relinquish unnecessary capacity
  void
  shrink_to_fit()
  {
    // TODO!
    // If the capacity is greater than the current size,
    //   "reserve" the exact amount of space needed to hold
    //   "size" elements.
    // capacity should be equal to size, and there should
    //   be no data loss
    // HINT: similar to reserve
  }

  // ////////// MODIFIERS //////////

private:

  iterator
  non_const_iterator (const_iterator pos)
  {
    return std::next (begin(), std::distance (cbegin(), pos));
  }

public:

  // Given an iterator, get a "safe" iterator.
  // An iterator is "safe" IFF there is enough capacity of the
  //   vector to add "elements" count to our vector without the
  //   need to perform a reallocation + move of data
  //
  // if size() + elements <= capacity(), convert the const
  //   iterator to a normal iterator (code provided)
  // if size() + elements > capacity(), you need to reallocate
  //   space for a larger data array. This will INVALIDATE all
  //   iterators -- so you need to (somehow) convert from a
  //   const_iterator to an index. HINT: std::distance()
  // If the capacity is insufficient, DOUBLE it.
  // If the capacity is 0, increase it to 1.
  iterator
  impl_safe_iterator (const_iterator pos, size_type elements = 1)
  {
    // TODO!
    // I've provided the implementation below to handle one of the cases
    return non_const_iterator (pos);
  }

  // removes all elements from the Vector
  void
  clear() noexcept
  {
    // TODO!
    // HINT: capacity shouldn't change but size should
  }

  // const-reference insert
  // returns an iterator to the newly inserted element
  iterator
  insert (const_iterator pos, T const& value)
  {
    iterator p = impl_safe_iterator (pos);
    std::copy_backward (p, end(), end() + 1);
    *p = value;
    ++m_size;
    return p;
  }

  // r-value reference insert
  // returns an iterator to the newly inserted element
  iterator
  insert (const_iterator pos, T&& value)
  {
    iterator p = impl_safe_iterator (pos);
    std::copy_backward (p, end(), end() + 1);
    *p = std::move (value);
    ++m_size;
    return p;
  }

  // size+value insert
  // returns an iterator to the first inserted element
  iterator
  insert (const_iterator pos, size_type count, T const& value)
  {
    iterator p = impl_safe_iterator (pos, count);
    std::copy_backward (p, end(), end() + count);
    std::fill_n (p, count, value);
    m_size += count;
    return p;
  }

  // Range insert
  // returns an iterator to the first inserted element
  template <typename InputIterator>
  std::enable_if_t<is_iterator<InputIterator>, iterator>
  insert (const_iterator pos, InputIterator first, InputIterator last)
  {
    size_type const count = std::distance (first, last);
    iterator p = impl_safe_iterator (pos, count);
    std::copy_backward (p, end(), end() + count);
    std::copy (first, last, p);
    m_size += count;
    return p;
  }

  // initializer_list insert
  // returns an iterator to the first inserted element
  iterator
  insert (const_iterator pos, std::initializer_list<T> ilist)
  {
    return insert (pos, ilist.begin(), ilist.end());
  }

  template <typename ... Args>
  iterator
  emplace (const_iterator pos, Args&& ... args)
  {
    iterator p = impl_safe_iterator (pos);
    std::copy_backward (p, end(), end() + 1);
    ++m_size;
    return new (p) T (std::forward<Args> (args)...);
  }

  // single-element erase. removes the element referred to by pos
  // returns an iterator to one-past-the-last element erased
  iterator
  erase (const_iterator pos)
  {
    // TODO!
    iterator p = non_const_iterator (pos);

    return p;
  }

  // range-base erase. removes all elements in the range [first, last)
  // returns an iterator to one-past-the-last element erased
  iterator
  erase (const_iterator first, const_iterator last)
  {
    // TODO!
    iterator fst = non_const_iterator (first);
    iterator lst = non_const_iterator (last);

    return fst;
  }

  // inserts a value to the end of the Vector
  void
  push_back (T const& value)
  {
    // TODO!
    // should be one line...
  }

  // inserts an r-value reference to the end of the Vector
  void
  push_back (T&& value)
  {
    // TODO! -- like above
    // HINT: when "using" value, say std::move(value)
  }

  // constructs an instance of T in-place at the end of the vector
  template <typename ... Args>
  reference
  emplace_back (Args&& ... args)
  {
    emplace (end(), std::forward<Args> (args)...);
    return back();
  }

  // Erase the element at the back.
  void
  pop_back()
  {
    // TODO!
    // should be one line...
  }

  // change the size of the vector
  void
  resize (size_type count)
  {
    // TODO!
    // If "count" is less than "size",
    //   erase the last elements.
    // If "count" is more than "size",
    //   insert T() at the end.
    // Change the size to be "count".
  }

  // change the size of the vector
  void
  resize (size_type count, value_type const& value)
  {
    // TODO!
    // Like above, but insert copies of value at
    //   the end if "count" is more than "size"
  }

  void
  swap (Vector& other)
  {
    std::swap (m_size, other.m_size);
    std::swap (m_capacity, other.m_capacity);
    std::swap (m_data, other.m_data);
  }

private:

  size_type m_size;
  size_type m_capacity;
  pointer m_data;
};

template <typename T>
bool
operator== (Vector<T> const& a, Vector<T> const& b) {
  return a.size() == b.size() && std::equal (a.begin(), a.end(), b.begin());
}

template <typename T>
bool
operator!= (Vector<T> const& a, Vector<T> const& b) {
  return a.size() != b.size() || !std::equal (a.begin(), a.end(), b.begin());
}

namespace std {

  template <typename T>
  void
  swap (Vector<T>& a, Vector<T>& b) {
    a.swap (b);
  }

}

#endif
